import click
import polars as pl

import yaml


class BlastProcessor:
    def __init__(self, blast, yaml, end_length=150, min_length=300):
        ## Initialise params
        self.end_length = end_length
        self.min_length = min_length

        self.adapter_yaml = None
        self.raw_blast = None
        self.hits = None
        self.actions = None
        self.bed = None

        ## Process blast hits
        try:
            ## Read raw data
            self.adapter_yaml = self._read_adapter_yaml(yaml)
            self.raw_blast = self._read_blast(blast)

            ## Validate inputs
            self._check_duplicate_adapter_matches(self.raw_blast, self.adapter_yaml)

            ## Determine hits
            self.hits = self._determine_blast_hits(self.raw_blast).collect()

            if self.hits.is_empty():
                click.echo(
                    "WARNING: After filtering, no BLAST hits remain. BED output will be empty."
                )
                return

            # Generate actions and BED output
            self.actions = self._determine_per_read_actions(self.hits.lazy()).collect()
            self.bed = self.generate_bed(self.actions.lazy()).collect()
        except pl.exceptions.NoDataError:
            click.echo(f"WARNING: {blast} was empty! BED output will be empty.")
            return

    def _read_adapter_yaml(self, yaml_path: str) -> pl.LazyFrame:
        """Open an adapter YAML file and return it as a lazy pl.DataFrame"""
        try:
            adapters = yaml.safe_load(yaml_path)
        except yaml.YAMLError as exc:
            print(exc)

        return pl.DataFrame(
            adapters,
            schema={
                "adapter": pl.String,
                "discard_middle": pl.Boolean,
                "discard_end": pl.Boolean,
                "trim_end": pl.Boolean,
                "middle_pident": pl.Float32,
                "end_pident": pl.Float32,
                "middle_length": pl.UInt8,
                "end_length": pl.UInt8,
            },
        ).lazy()

    def _read_blast(self, blast_path: str) -> pl.LazyFrame:
        """Read BLAST outformat file to a LazyFrame.

        If a BAM file is provided and there is no 13th column (containing read lengths)
        then these can be quickly counted.

        Returns with the columns qseqid, sseqid, pident, length, qstart, qend, and evalue.
        qstart and qend are sorted to ignore information about strand.
        """
        blastout = pl.scan_csv(
            blast_path,
            has_header=False,
            separator="\t",
            schema={
                "qseqid": pl.String,
                "sseqid": pl.String,
                "pident": pl.Float32,
                "length": pl.UInt8,
                "mismatch": pl.UInt8,
                "gapopen": pl.UInt8,
                "qstart": pl.UInt32,
                "qend": pl.UInt32,
                "sstart": pl.UInt8,
                "send": pl.UInt8,
                "evalue": pl.Float32,
                "bitscore": pl.Float32,
                "read_length": pl.UInt32,
            },
        ).with_columns(
            pl.col("qseqid")
            .cast(pl.Categorical)
            .set_sorted(),  ## Set sorted for faster grouping operations
            pl.col("sseqid").cast(pl.Categorical),
            pl.when(pl.col("qstart") > pl.col("qend"))
            .then(
                pl.struct(
                    start="qend",
                    end="qstart",
                )
            )
            .otherwise(pl.struct(["qstart", "qend"]))
            .struct.field(["qstart", "qend"]),
        )

        return blastout

    def _check_duplicate_adapter_matches(
        self, blast: pl.LazyFrame, adapter_df: pl.DataFrame
    ) -> list:
        """Check if any adapters in a BLAST dataframe match to more than one
        adapter sequence in the YAML file.
        """
        counts = (
            blast.select(pl.col("sseqid"))
            .unique()
            .with_columns(pl.col("sseqid").cast(pl.String))
            .join_where(adapter_df, pl.col("sseqid").str.contains(pl.col("adapter")))
            .group_by("sseqid")
            .len()
            .filter(pl.col("len") > 1)
            .collect()["sseqid"]
            .to_list()
        )

        if len(counts) > 0:
            raise ValueError(
                f"Adapters {', '.join(counts)} in the BLAST table match to more than one adapter defined in the YAML!"
            )

    def _determine_blast_hits(self, blast: pl.LazyFrame) -> pl.LazyFrame:
        """Determine whether each blast hit matches sufficient criteria for discard/trimming

        Keyword arguments:
        blastout: pl.DataFrame containing the results of a blast search with read lengths in column 13
        adapter_key: pl.DataFrame of adapters and their filtering/trimming parameters
        end_length: length of window at either end of read which is examined for trimming
        """
        ## When to discard a read due to adapter presence
        discard = (
            (pl.col("discard_middle")).and_(
                (pl.col("read_length") > 2 * self.end_length),
                (pl.col("qstart") > self.end_length),
                (pl.col("qend") < pl.col("read_length") - self.end_length),
                (pl.col("qend") > self.end_length),
                (pl.col("qstart") < pl.col("read_length") - self.end_length),
                (pl.col("pident") >= pl.col("middle_pident")),
                (pl.col("length") >= pl.col("middle_length")),
            )
        ) | (
            (pl.col("discard_end")).and_(
                (
                    (pl.col("qend") < self.end_length)
                    | (pl.col("qstart") > pl.col("read_length") - self.end_length)
                ),
                (pl.col("pident") >= pl.col("end_pident")),
                (pl.col("length") >= pl.col("end_length")),
            )
        )

        ## When to trim a read at the left end
        trim_l = (pl.col("trim_end")).and_(
            pl.col("qend") < self.end_length,
            (pl.col("pident") >= pl.col("end_pident")),
            (pl.col("length") >= pl.col("end_length")),
        )

        ## When to trim a read at the righr end
        trim_r = (pl.col("trim_end")).and_(
            pl.col("qstart") >= pl.col("read_length") - self.end_length,
            (pl.col("pident") >= pl.col("end_pident")),
            (pl.col("length") >= pl.col("end_length")),
        )

        return (
            blast
            ## match each blast hit with an adapter
            .join_where(
                self.adapter_yaml,
                pl.col("sseqid").cast(pl.String).str.contains(pl.col("adapter")),
            )
            ## Determine which hits to keep and return rows with a "real" hit
            .with_columns(discard=discard, trim_l=trim_l, trim_r=trim_r)
            .filter(pl.any_horizontal("discard", "trim_l", "trim_r"))
        )

    def _determine_per_read_actions(self, hits: pl.LazyFrame) -> pl.LazyFrame:
        """Take a dataframe of matches from match_hits()
        and process the result to decide which actions to take.

        Keyword arguments:
        hits: pl.DataFrame of blast hits with matches
        end_length: length of window at either end of read which is examined for trimming
        min_length: minimum length of a read after trimming to keep
        """
        hits_actions = hits.with_columns(
            (
                pl.col("read_length")
                - ((pl.col("trim_l").any() + pl.col("trim_r").any()) * self.end_length)
            )
            .over("qseqid")
            .alias("read_length_after_trimming"),
        ).with_columns(
            gen_action=(
                pl.when(
                    pl.col("discard").any()
                    | (pl.col("read_length_after_trimming") < self.min_length)
                )
                .then(pl.lit("discard"))
                .otherwise(pl.lit("trim"))
                .over("qseqid")
            )
        )

        discard = (
            hits_actions.filter(pl.col("gen_action") == "discard")
            .group_by("qseqid")
            .agg(
                pl.col("sseqid", "qstart", "qend", "read_length").get(
                    pl.col("evalue").arg_min()
                ),
                action=pl.lit("discard"),
            )
        )

        trim_l = (
            hits_actions.filter(~(pl.col("gen_action") == "discard"), pl.col("trim_l"))
            .group_by("qseqid")
            .agg(
                pl.col("sseqid", "qstart", "qend", "read_length")
                .filter(pl.col("qend") <= self.end_length)
                .get(
                    pl.col("evalue").filter(pl.col("qend") <= self.end_length).arg_min()
                ),
                action=pl.lit("trim_l"),
            )
        )

        trim_r = (
            hits_actions.filter(~(pl.col("gen_action") == "discard"), pl.col("trim_r"))
            .group_by("qseqid")
            .agg(
                pl.col("sseqid", "qstart", "qend", "read_length")
                .filter(pl.col("qstart") >= pl.col("read_length") - self.end_length)
                .get(
                    pl.col("evalue")
                    .filter(pl.col("qstart") >= pl.col("read_length") - self.end_length)
                    .arg_min()
                ),
                action=pl.lit("trim_r"),
            )
        )

        return pl.concat([discard, trim_l, trim_r])

    def generate_bed(self, actions) -> pl.LazyFrame:
        """Format the output from determine_actions() into valid BED."""
        return actions.select(
            qseqid=pl.col("qseqid"),
            start=pl.when(
                (pl.col("action") == "discard") | (pl.col("action") == "trim_l")
            )
            .then(pl.lit(0))
            .when(pl.col("action") == "trim_r")
            .then(pl.col("read_length") - self.end_length)
            .otherwise(pl.lit(0)),
            end=pl.when(
                (pl.col("action") == "discard") | (pl.col("action") == "trim_r")
            )
            .then(pl.col("read_length"))
            .when(pl.col("action") == "trim_l")
            .then(pl.lit(self.end_length))
            .otherwise(pl.lit(0)),
            reason=pl.concat_str(pl.col("action"), pl.lit(":"), pl.col("sseqid")),
        ).sort("qseqid", "start")

    def generate_hits(self) -> pl.LazyFrame:
        return self.hits.select(
            [
                "qseqid",
                "sseqid",
                "pident",
                "length",
                "mismatch",
                "gapopen",
                "qstart",
                "qend",
                "sstart",
                "send",
                "evalue",
                "bitscore",
                "read_length",
                "discard",
                "trim_l",
                "trim_r",
            ]
        )



import bgzip
import click
import polars as pl
import pysam


class BamFilter:
    def __init__(
        self, bam: click.File, bed: str, outfile: str, threads: int, fastq: bool
    ):
        self.bam = bam
        self.bed = bed
        self.outfile = outfile
        self.threads = threads
        self.write_fastq = fastq

    def format_fastx_record(self, header: str, sequence: str, qual: str) -> str:
        """Format a header and sequence into FASTA format"""
        if qual is not None:
            return f"@{header}\n{sequence}\n+\n{qual}\n"
        else:
            return f">{header}\n{sequence}\n"

    def trim_positions(self, seq: str, ranges: list) -> str:
        """Trim DNA sequence seq to remove the positions specified in ranges.

        seq: string
        ranges: list of non-overlapping tuples with (start, end)
        """
        ## sort the ranges so we trim from the end - that way indexing stays constant
        ranges = sorted(ranges, key=lambda x: x[0], reverse=True)

        for start, end in ranges:
            seq = seq[:start] + seq[end:]

        return seq

    def filter_bam_with_bed(self) -> iter:
        try:
            bed_df = pl.read_csv(
                self.bed,
                separator="\t",
                has_header=False,
                schema={
                    "read": pl.String,
                    "start": pl.Int64,
                    "end": pl.Int64,
                    "reason": pl.String,
                },
            )
            filters = bed_df.iter_rows(named=True)

            r = next(filters, None)
        except pl.exceptions.NoDataError:
            click.echo("WARN: BED file is empty! Reads will be streamed as-is.")
            filters = iter([])
            r = next(filters, None)

        with bgzip.BGZipWriter(self.outfile, num_threads=self.threads) as out:
            with pysam.AlignmentFile(
                self.bam, "rb", check_sq=False, require_index=False
            ) as b:
                ## Process: for each read in the BAM, check if it matches the current BED record.
                ## If yes, pull BED records until we reach a record for the next read.
                ## Then trim the sequence using the records pulled and write to fasta.
                ## If not, write record straight to disk
                for read in b.fetch(until_eof=True):
                    if r is not None and r["read"] == read.query_name:
                        ranges = [(int(r["start"]), int(r["end"]))]

                        while True:
                            r = next(filters, None)
                            if r is None or r["read"] != read.query_name:
                                break
                            ranges.append((int(r["start"]), int(r["end"])))

                        sequence = self.trim_positions(read.query_sequence, ranges)

                        qual = None
                        if self.write_fastq:
                            qual = self.trim_positions(read.qual, ranges)

                        print(
                            f"Processing read: {read.query_name}, ranges: {ranges}, original length: {read.query_length}, new_length: {len(sequence)}"
                        )
                        if len(sequence) > 0:
                            out.write(
                                self.format_fastx_record(
                                    read.query_name, sequence, qual
                                ).encode("utf-8")
                            )
                    else:
                        qual = None
                        if self.write_fastq:
                            qual = read.qual

                        out.write(
                            self.format_fastx_record(
                                read.query_name, read.query_sequence, qual
                            ).encode("utf-8")
                        )

        try:
            read = next(filters)
            print(f"Read {read['read']} not processed!")
            for read in filters:
                print(f"Read {read['read']} not processed!")

            raise RuntimeError(
                "ERROR: Not all entries in the BED file were processed! Are they sorted in the same order as the BAM file?"
            )
        except StopIteration:
            print("Read filtering complete!")



import bgzip
import click
import json
import re

from hifi_trimmer.BamFilter import BamFilter
from hifi_trimmer.BlastProcessor import BlastProcessor
from hifi_trimmer.SummariseBlastResults import SummariseBlastResults


@click.group()
@click.version_option(package_name="hifi_trimmer")
def cli():
    """Main entry point for the tool."""
    pass


@click.command("process_blast")
@click.argument("blastout", type=click.Path(exists=True))
@click.argument("adapter_yaml", type=click.File(mode="r"))
@click.option(
    "-p",
    "--prefix",
    default=None,
    required=False,
    type=str,
    help="Output prefix for results. Defaults to the basename of the blastout if not provided.",
)
@click.option(
    "-ml",
    "--min_length_after_trimming",
    default=300,
    show_default=True,
    type=int,
    help="Minumum length of a read after trimming the ends in order not to be discarded",
)
@click.option(
    "-el",
    "--end_length",
    default=150,
    show_default=True,
    type=int,
    help="Window size at either end of the read to be considered as 'ends' for searching",
)
@click.option(
    "-hf",
    "--hits",
    "hits_flag",
    default=False,
    is_flag=True,
    type=bool,
    help="""Write the hits identified using the given adapter specifications to TSV. The format is
    standard BLAST outfmt 6 with the following extra columns: read_length (int), discard (bool), trim_l (bool), trim_r (bool)
    """,
)
@click.option(
    "--no-summary",
    default=False,
    type=bool,
    is_flag=True,
    help="Skip writing a summary JSON with the number of hits for each adapter",
)
@click.option(
    "-t",
    "--threads",
    default=1,
    show_default=True,
    type=int,
    help="Number of threads to use for compression",
)
def process_blast(
    blastout: str,
    adapter_yaml: click.File,
    prefix: str,
    min_length_after_trimming: int,
    end_length: int,
    hits_flag: bool,
    no_summary: bool,
    threads: int,
) -> None:
    """Processes the input blastout file according to the adapter yaml key.

    BLASTOUT: tabular file resulting from a BLAST query of a readset against a BLAST
    database of adapter sequences, run with -outfmt "6 std qlen".

    ADAPTER_YAML: yaml file containing a list with the following fields per adapter:

    \b
    - name: (name of adapter. can be a regular expression)
      discard_middle: True/False (discard read if adapter found in middle)
      discard_end: True/False (discard read if adapter found in end)
      trim_end: True/False (trim read if adapter found in end)
      middle_pident: int (minimum pident requred to identify adapter in middle of read)
      middle_length: int (minimum match length required to identify adapter in middle of read)
      end_pident: int (minimum pident requred to identify adapter in end window)
      end_length: int (minimum match length requred to identify adapter in end window)

    Output:
    By default, writes bgzipped BED to [prefix].bed.gz, and a JSON summary file with raw counts
    of adapter hits detected, counts identified after processing, and the total length of
    removed sequences per adapter to [prefix].summary.json.
    """
    if prefix is None:
        prefix = re.sub("\\.blastout.*$", "", blastout)

    ## Process BLAST inputs
    results = BlastProcessor(
        blastout, adapter_yaml, end_length, min_length_after_trimming
    )

    ## Write output hits if requested
    if hits_flag and results.hits is not None and not results.hits.is_empty():
        results.generate_hits().write_csv(
            prefix + ".hits", separator="\t", include_header=False
        )

    ## Write BED file
    click.echo(f"Writing BED file: {prefix}.bed.gz")
    out_bed = (
        results.bed.write_csv(separator="\t", include_header=False).encode("utf-8")
        if results.bed is not None
        else b""
    )
    with open(prefix + ".bed.gz", "wb") as f:
        with bgzip.BGZipWriter(f, num_threads=threads) as out:
            out.write(out_bed)

    if not no_summary:
        summary = SummariseBlastResults(results)
        with open(prefix + ".summary.json", "w") as f:
            json.dump(summary.generate_summary(), f, indent=4)


@click.command("filter_bam")
@click.argument("bam", type=click.File(mode="rb"))
@click.argument("bed", type=click.Path(exists=True))
@click.argument("outfile", default=None, required=True, type=click.File(mode="wb"))
@click.option(
    "-f",
    "--fastq",
    is_flag=True,
    default=False,
    help="Write FASTQ instead of FASTA",
)
@click.option(
    "-t",
    "--threads",
    default=1,
    show_default=True,
    type=int,
    help="Number of threads to use for compression",
)
def filter_bam(
    bam: click.File, bed: str, outfile: click.File, threads: int, fastq: bool
) -> None:
    """
    Filter the reads stored in a BAM file using the appropriate BED file produced
    by blastout_to_bed and write to a bgzipped fasta file.

    \b
    BAM: BAM file in which to filter reads
    BED: BED file describing regions of the read set to exclude.
    OUTFILE: File to write the filtered reads to (bgzipped).
    """
    click.echo(f"Filtering {bam} using BED file: {bed}")
    click.echo(f"Writing the output to {outfile}.")
    filterer = BamFilter(bam, bed, outfile, threads, fastq)
    filterer.filter_bam_with_bed()


cli.add_command(process_blast)
cli.add_command(filter_bam)

if __name__ == "__main__":
    cli()



from click.testing import CliRunner
from hifi_trimmer.hifi_trimmer import cli
import pathlib
import hashlib
import pytest


def list_test_data(test_dir):
    """
    Get a list of files in the test directory and store them in
    a dict indexed by extension.
    """
    test_data = {}
    for p in pathlib.Path(test_dir).iterdir():
        if p.suffix == ".gz":
            ext = p.suffixes[-2][1:]
        else:
            ext = p.suffix[1:]

        test_data[ext] = str(p.absolute())

    return test_data


def list_test_dirs():
    """
    List all the data dirs in the test data directory.
    """
    data_dir = pathlib.Path(__file__).parent / "data"

    data = []
    for p in pathlib.Path(data_dir).iterdir():
        if p.is_dir():
            data.append(list_test_data(p))

    return data


def md5checksum(fname):
    """
    Get the MD5 checksum for a file.
    """
    md5 = hashlib.md5()
    with open(fname, "rb") as f:
        while chunk := f.read(4096):
            md5.update(chunk)

    return md5.hexdigest()


@pytest.mark.parametrize("testdata", list_test_dirs())
def test_process_blast(tmp_path, testdata):
    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as td:
        result = runner.invoke(
            cli,
            [
                "process_blast",
                "--prefix",
                "test",
                testdata["blastout"],
                testdata["yaml"],
            ],
        )

        print(result.stdout)

        md5_bed = md5checksum(pathlib.Path(td) / "test.bed.gz")
        md5_summary = md5checksum(pathlib.Path(td) / "test.summary.json")

    assert result.exit_code == 0
    assert md5_bed == md5checksum(testdata["bed"])
    assert md5_summary == md5checksum(testdata["json"])


@pytest.mark.parametrize("testdata", list_test_dirs())
def test_filter_bam(tmp_path, testdata):
    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as td:
        result = runner.invoke(
            cli,
            ["filter_bam", testdata["bam"], testdata["bed"], "test.filtered.fa.gz"],
        )

        md5_fasta = md5checksum(pathlib.Path(td) / "test.filtered.fa.gz")

    assert result.exit_code == 0
    assert md5_fasta == md5checksum(testdata["fa"])


@pytest.mark.parametrize("testdata", list_test_dirs())
def test_filter_bam_fastq(tmp_path, testdata):
    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as td:
        result = runner.invoke(
            cli,
            [
                "filter_bam",
                testdata["bam"],
                testdata["bed"],
                "test.filtered.fq.gz",
                "--fastq",
            ],
        )

        md5_fasta = md5checksum(pathlib.Path(td) / "test.filtered.fq.gz")

    assert result.exit_code == 0
    assert md5_fasta == md5checksum(testdata["fq"])



import click
import polars as pl
import polars.selectors as cs

from hifi_trimmer.BlastProcessor import BlastProcessor


class SummariseBlastResults:
    def __init__(self, result: BlastProcessor):
        self.blast = result.raw_blast
        self.hits = result.hits
        self.actions = result.actions
        self.end_length = result.end_length

    def _summarise_blast(self, blast) -> pl.DataFrame:
        """
        Summarise a raw BLAST table, returning the number of hits by
        adapter.
        """
        return (
            self.blast.group_by("sseqid")
            .len("n_hits")
            .rename({"sseqid": "adapter"})
            .sort("adapter")
            .collect()
        )

    def _summarise_hits(self, hits) -> pl.DataFrame:
        """
        Summarise a filtered BLAST table, returning the number of hits by
        adapter and action.
        """
        return (
            self.hits.select(
                sseqid=pl.col("sseqid"),
                discard=pl.col("discard"),
                trim=(pl.col("trim_l") + pl.col("trim_r")).cast(pl.Boolean),
            )
            .unpivot(
                cs.by_name(["discard", "trim"]),
                index="sseqid",
                variable_name="action",
            )
            .filter(pl.col("value"))
            .group_by(["sseqid", "action"])
            .len(name="n_hits")
            .filter(pl.col("n_hits") > 0)
            .rename({"sseqid": "adapter"})
            .sort(["adapter", "action"])
        )

    def _summarise_adapter_actions(self, actions) -> pl.DataFrame:
        """
        Summarise an actions table, returning the number of reads affected
        and the number of bases removed for each adapter and action.
        """
        return (
            self.actions.with_columns(
                bases=pl.when(pl.col("action") == "discard")
                .then(pl.col("read_length"))
                .when(pl.col("action").str.contains("trim"))
                .then(pl.lit(self.end_length)),
                action=pl.when(pl.col("action") == "discard")
                .then(pl.lit("discard"))
                .otherwise(pl.lit("trim")),
            )
            .group_by(["sseqid", "action"])
            .agg(
                n_reads=pl.col("qseqid").unique().len(),
                bases_removed=pl.col("bases").sum(),
            )
            .filter(pl.col("n_reads") > 0)
            .rename({"sseqid": "adapter"})
            .sort(["adapter", "action"])
        )

    def _summarise_actions(self, actions) -> pl.DataFrame:
        """
        Summarise an actions table, returning the total number of reads affected
        and the number of bases removed for each action.
        """
        return (
            self.actions.with_columns(
                bases=pl.when(pl.col("action") == "discard")
                .then(pl.col("read_length"))
                .when(pl.col("action").str.contains("trim"))
                .then(pl.lit(self.end_length)),
                action=pl.when(pl.col("action") == "discard")
                .then(pl.lit("discard"))
                .otherwise(pl.lit("trim")),
            )
            .group_by(["action"])
            .agg(
                n_reads=pl.col("qseqid").unique().len(),
                bases_removed=pl.col("bases").sum(),
            )
        )

    def generate_summary(self) -> dict:
        """
        Takes the raw BLAST table, the BLAST table filtered for "valid" hits
        using criteria from the YAML, and the per-read "actions" table,
        and returns a dictionary with a summary about the adapter removal process:

        detected: how many hits per adapter detected in the raw BLAST output
        hits: for each adapter and action, the following stats:
            - number of blast hits
            - number of reads where this was the best hit
            - number of bases removed
        total_reads_discarded: total number of reads discarded
        total_reads_trimmed: total number of reads trimmed
        total_bases_removed: total length of removed bases
        """
        summary = {
            "detections": [],
            "hits": [],
            "total_reads_discarded": 0,
            "total_reads_trimmed": 0,
            "total_bases_removed": 0,
        }

        try:
            blast_summary = self._summarise_blast(self.blast)
            hits_summary = self._summarise_hits(self.hits)

            if not blast_summary.is_empty():
                summary["detections"] = blast_summary.to_dicts()

                if not hits_summary.is_empty():
                    adapter_actions_summary = self._summarise_adapter_actions(
                        self.actions
                    )
                    all_actions_summary = self._summarise_actions(self.actions)

                    summary["hits"] = (
                        hits_summary.join(
                            adapter_actions_summary, on=["adapter", "action"]
                        )
                        .sort(["adapter", "action"])
                        .to_dicts()
                    )

                    summary["total_bases_removed"] = all_actions_summary[
                        "bases_removed"
                    ].sum()
                    summary["total_reads_discarded"] = all_actions_summary.filter(
                        pl.col("action") == "discard"
                    )["n_reads"].sum()
                    summary["total_reads_trimmed"] = all_actions_summary.filter(
                        pl.col("action") == "trim"
                    )["n_reads"].sum()

        except pl.exceptions.NoDataError:
            click.echo("Writing empty summary file!")

        return summary
